/**************************************************************
 * File: agrsoftmodem.c
 *
 * Copyright (c) 2002, 2003, 2004, 2005 Agere Systems, Inc.  All rights reserved.
 *
 * Description:
 *   Agere Soft Modem driver interface functions for Linux
 *
 * Revision History:
 *   Name                   Date          Change
 *   Soumyendu Sarkar       12/03/2002    Initial as agrmodem.c
 *   Soumyendu Sarkar       03/11/2004    Adapted to kernel 2.6.x with PCI support
 *   Soumyendu Sarkar       11/16/2005    Adapted to kernel 2.6.14 with PCI PP modem support
 *************************************************************/

/*
 * Modified by A.Chentsov
 */
 
#define LUCENT_MODEM
#ifndef AGERE_SOFT_MODEM
#define AGERE_SOFT_MODEM
#endif

#include <linux/module.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION (2, 6, 19)
# include <linux/config.h>
#endif

#include <linux/init.h>

#include <linux/sched.h>

#include <linux/ctype.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/time.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/smp.h>
#include <asm/io.h>
#include <asm/atomic.h>
#include "linuxif.h"
#include "agrmodem.h"

#include "agr.h"


#define AGRSM_PREFIX	"agrsm"
#define AGRSM_INFO(fmt...)           printk (KERN_INFO AGRSM_PREFIX ": " fmt)

extern unsigned int BaseAddress;

#ifndef LT_KER_26
#define LT_KER_26 1
#endif

int numSupportedDevice;
/** list of supported devices - vendorID, deviceID, subSystemID, subVendorID **/
/* if both subSystemID and subVendorID are set to 0, the driver will ignore these fields */
const SUPPORTED_DEVICES agereSupDevices[] = {
  { 0x11C1, 0x0620, 0x0000, 0x0000 },
  { 0x11C1, 0x0630, 0x0000, 0x0000 },
  { 0x8086, 0x2416, 0x0000, 0x0000 },
  { 0x8086, 0x2426, 0x0000, 0x0000 },
  { 0x8086, 0x2446, 0x0000, 0x0000 },
  { 0x8086, 0x7196, 0x0000, 0x0000 },
  { 0x8086, 0x2486, 0x0000, 0x0000 },
  { 0x8086, 0x24C6, 0x0000, 0x0000 },
  { 0x8086, 0x24D6, 0x0000, 0x0000 },
  { 0x11C1, 0x048C, 0x0000, 0x0000 },
  { 0x11C1, 0x048F, 0x0000, 0x0000 },
  { 0x8086, 0x266D, 0x0000, 0x0000 },
};

/************** Ensure Kernel Version 2.4.0 or later *************/
#if LINUX_VERSION_CODE < (((2)<<16)|((4)<<8)|((0)))
#error This driver compiles for Linux versions 2.4.0 and later
#endif

static char *modem_name = "Agere Modem Controller driver";
static char *modem_version = "2.1.60.0.Linspire";
static char *modem_revdate = "2005-11-16";
extern void x_debug_msg (char *);
struct timer_list timerList;
int (*agr_rs_interrupt)(void) = NULL;
spinlock_t modem_driver_lock = SPIN_LOCK_UNLOCKED;

/********************* Interface Operations *********************/
int GetAgrModemInterface(struct _agrsm_ops *agrsm_ops)
{
	agrsm_ops->detect_modem = wrap_lucent_detect_modem;
	agrsm_ops->init_modem = wrap_lucent_init_modem;
	agrsm_ops->PortOpen = modemPortOpen;
	agrsm_ops->PortClose = modemPortClose;
	agrsm_ops->read_vuart_register = wrap_read_vuart_port;
	agrsm_ops->write_vuart_register = wrap_write_vuart_port;
	agrsm_ops->app_ioctl_handler = wrap_app_ioctl_handler;
	agrsm_ops->dsp_isr = wrap_dum_dp_dsp_isr;
	agrsm_ops->read_buffer = wrap_V16550_Read_RBR_buffer;
	agrsm_ops->rs_interrupt_ptr = &agr_rs_interrupt;

	return TRUE;
}

EXPORT_SYMBOL (GetAgrModemInterface);


/************************ Open / Close *************************/
int modemPortOpen ( void )
{
	wrap_linux_modem_open ();
	return( wrap_vxdPortOpen() );
}

void modemPortClose ( void )
{
	wrap_vxdPortClose();
	agr_rs_interrupt = NULL;
	wrap_linux_modem_close();
}


/* agrsm_devprobe:
 *  fill info in agrInfos
 *  to be passed later to LXHardwareInfoCreate ...
 */
static int __devinit 
agrsm_devprobe (struct pci_dev *dev, const struct pci_device_id *ent)
{
	// int idx = agrInfos.count;
	agr_instance_info_t *info = &agrInfos.slot[agrInfos.count];
	int bar, ret;
	extern u32 BaseAddress, BaseAddress2;
	extern u8 Irq;

	ret = pci_enable_device (dev);
	if (ret != 0) 
		return -ENODEV;

	if (agrInfos.count == 0x20)
		return -ENODEV;

	/* resources, type */
	info->hierarchy_name = "AgerePCI";
	switch (dev->device) {
	case 0x48c:
		info->name = "Perseus";
		info->codec_type = 0x2b;
		break;
		
	case 0x620:
		info->name = "PP";
		info->codec_type = 0x33;
		break;
	}

	info->spec2 = 0;
	info->res.cd = 1;
	info->res.irq = dev->irq;
	info->res.io = 1;

	//info->res.res0;
	for (bar = 5; bar >= 0; bar--) 
		if ( pci_resource_start (dev, bar) != 0 && 
			(pci_resource_flags (dev, bar) & IORESOURCE_IO) ) {
			info->res.res0 =  pci_resource_start (dev, bar);
			break;
		}

	/*if (bar < 0) {
	} */

	info->num = 1; /* all pci devs - single nodes */
	info->node = 0;
	info->all_idx = agrInfos.count;
	info->g_idx = agrInfos.count;

	agrInfos.count++;

	BaseAddress = info->res.res0;
	BaseAddress2 = 0;
	Irq = info->res.irq;

	return 0;
}

asmlinkage int LXHardwareInfoDestroy (pagr_info_t *ppinfo)
{ return 0;}

/* override LXHardwareInfoCreate */
asmlinkage int LXHardwareInfoCreate (pagr_info_t *ppinfo)
{
	*ppinfo = &agrInfos; /* provide prepared infos */

	return 0;
}


static void __devexit agrsm_remove (struct pci_dev *dev) 
{
	pci_disable_device (dev);
}

__devinitdata static struct pci_device_id agrsm_ids[] = {
	{ PCI_DEVICE (PCI_VENDOR_ID_ATT, 0x48c) },
	{ PCI_DEVICE (PCI_VENDOR_ID_ATT, 0x620) },
	{ 0 }
};

struct pci_driver agrsm_drv = {
	.name		= "agrsm",
	.probe		= agrsm_devprobe,
	.remove		= __devexit_p (agrsm_remove),
	.id_table	= agrsm_ids
};


/************************** Entry Points ***************************/
int __init modem_init_module (void)
{
	int ret = 0;
	numSupportedDevice = sizeof(agereSupDevices) / sizeof(agereSupDevices[0]);

	memset (&agrInfos, 0, sizeof agrInfos);
	agrInfos.count = 0;

#if   LINUX_VERSION_CODE  >=  KERNEL_VERSION (2, 6, 10)
	ret = pci_register_driver (&agrsm_drv);
#else
	ret = pci_module_init (&argsm_drv);
#endif

	if (ret)
		return ret;
		
	AGRSM_INFO ("created %d infos\n", agrInfos.count);

	ret = wrap_linux_modem_init ();
	if (!ret) {
		printk("<1>Loading module %s version %s (%s)\n", modem_name, modem_version, modem_revdate);
	}
	else
		printk("<1>Could not detect Agere soft modem device\n Agere soft modem driver not loaded\n");

	return (ret);
}

void __exit modem_cleanup_module (void)
{
	wrap_linux_modem_cleanup ();

	pci_unregister_driver (&agrsm_drv);

	printk("<1>Unloading %s: version %s\n", modem_name, modem_version);
}

module_init(modem_init_module);
module_exit(modem_cleanup_module);
MODULE_AUTHOR("Agere Systems Inc");
MODULE_DESCRIPTION("Agere Modem Controller driver");
MODULE_LICENSE("Proprietary");

/************************** IO - Interrupts ************************/
fnatr void *x_request_region(unsigned long start, unsigned long n, const char *name)
{
//  return(request_region(start, n, name));
#if LINUX_VERSION_CODE < KERNEL_VERSION (2, 6, 29)
  return(__request_region(&ioport_resource, (start), (n), (name)));
#else
  return(__request_region(&ioport_resource, (start), (n), (name), 0));
#endif
}

#ifndef LT_KER_26
int x_check_region(unsigned long start, unsigned long n) { return (check_region (start, n)); }
void x_release_region(unsigned long start, unsigned long n) { release_region(start, n); }
#else
fnatr int x_check_region(unsigned long start, unsigned long n) { return (__check_region(&ioport_resource, start, n)); }
fnatr void x_release_region(unsigned long start, unsigned long n) { __release_region(&ioport_resource, (start), (n)); }
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION (2, 6, 19)
static irqreturn_t wrap_LX_isr_handler (int irq, void *_isr, struct pt_regs *regs) {
	LX_isr_handler (irq, _isr, regs); 
	return IRQ_HANDLED; /* TODO */
}
#else
static irqreturn_t wrap_LX_isr_handler (int irq, void *_isr) {
	LX_isr_handler (irq, _isr, NULL); 
	return IRQ_HANDLED; /* TODO */
}
#endif

asmlinkage int x_request_irq(unsigned int irq, void (*handler), unsigned long irqflags, const char * devname, void *dev_id)
{
	return (request_irq(irq, wrap_LX_isr_handler, irqflags, devname, dev_id));
}
fnatr void x_free_irq(unsigned int irq, void *dev_id) { free_irq( irq, dev_id); }

/********************** Time Operations *************************/
fnatr void x_ms_delay (unsigned ms) { mdelay (ms); }
fnatr void *x_do_gettimeofday(void *tval) { do_gettimeofday ((struct timeval *)tval); return(tval); }
fnatr void lt_add_timer( void (*timerfunction)(unsigned long) )
{
	timerList.expires = jiffies+1;
	timerList.function = wrap_timertick_function;
	timerList.data = 0;
	add_timer(&timerList);
}
fnatr dword VMODEM_Get_System_Time (void)
{
	struct timeval time;
	do_gettimeofday( &time);
	return((time.tv_usec/1000)+(time.tv_sec*1000));
}
fnatr void lt_init_timer ( void  ) { init_timer(&timerList); }

/************************** OS Operations **************************/
fnatr int x_smp_processor_id (void) { return (smp_processor_id ()); }
#ifndef LT_KER_26
int x_smp_num_cpus (void) {if (smp_num_cpus > 0) return smp_num_cpus;else return 1; }
#else
fnatr int x_smp_num_cpus (void) {if (num_online_cpus() > 0) return (num_online_cpus());else return 1; }
#endif

#ifndef LT_KER_26
void x_MOD_INC_USE_COUNT (void) { MOD_INC_USE_COUNT; }
void x_MOD_DEC_USE_COUNT (void) { MOD_DEC_USE_COUNT; }
#else
fnatr void x_MOD_INC_USE_COUNT (void) { try_module_get(THIS_MODULE); }
fnatr void x_MOD_DEC_USE_COUNT (void) { module_put(THIS_MODULE); }
#endif

/* called by kernel */

#if LINUX_VERSION_CODE < KERNEL_VERSION (2, 6, 20)
static void wrap_LXHardwareBottomHalf (void *hardware) {
	LXHardwareBottomHalf (hardware);
}
#else
static void wrap_LXHardwareBottomHalf (struct work_struct *work) {
	agr_hardware_t *hw = container_of (work, agr_hardware_t, work);
	LXHardwareBottomHalf (hw);
}
#endif

/* called by core */
//asmlinkage void x_task_queue_init (x_tq_struct_t *task_x, x_func_ptr func, void *data);
//void x_queue_task (x_tq_struct_t *task_x) fnatr;

asmlinkage void x_task_queue_init (x_tq_struct_t *task_x, x_func_ptr func, void *data)
	{
#ifndef LT_KER_26
	struct tq_struct *x_tqueue = (struct tq_struct *) task_x;
	INIT_TQUEUE(x_tqueue, func, data);
#else
	struct work_struct *x_tqueue = (struct work_struct *) task_x;
# if LINUX_VERSION_CODE < KERNEL_VERSION (2, 6, 20)
	INIT_WORK(x_tqueue, wrap_LXHardwareBottomHalf, data);
# else
	INIT_WORK(x_tqueue, wrap_LXHardwareBottomHalf);
# endif 
#endif
	}
fnatr void x_queue_task	(x_tq_struct_t *task_x)
	{
#ifndef LT_KER_26
	struct tq_struct *x_tqueue = (struct tq_struct *) task_x;
	queue_task (x_tqueue, &tq_immediate);
	mark_bh (IMMEDIATE_BH);
#else
//# if LINUX_VERSION_CODE < LINUX_KERNEL (2, 6, 20)
	struct work_struct *x_tqueue = (struct work_struct *) task_x;
	schedule_work(x_tqueue);
#endif
}
